import torch
import pandas as pd
from PIL import Image
import io
from torch import nn
from torch.utils.data import Dataset, DataLoader
import torchvision.transforms as transforms
from aiia.model import AIIABase
from sklearn.model_selection import train_test_split
from typing import Dict, List, Union


class ImageDataset(Dataset):
    def __init__(self, dataframe, transform=None):
        self.dataframe = dataframe
        self.transform = transform

    def __len__(self):
        return len(self.dataframe)

    def __getitem__(self, idx):
        row = self.dataframe.iloc[idx]

        # Decode image_512 from bytes
        img_bytes = row['image_512']
        img_stream = io.BytesIO(img_bytes)
        low_res_image = Image.open(img_stream).convert('RGB')

        # Decode image_1024 from bytes
        high_res_bytes = row['image_1024']
        high_stream = io.BytesIO(high_res_bytes)
        high_res_image = Image.open(high_stream).convert('RGB')

        # Apply transformations if specified
        if self.transform:
            low_res_image = self.transform(low_res_image)
            high_res_image = self.transform(high_res_image)

        return {'low_res': low_res_image, 'high_res': high_res_image}




class ModelTrainer:
    def __init__(self,
                 model_name: str = "AIIA-Base-512",
                 dataset_paths: List[str] = None,
                 batch_size: int = 32,
                 learning_rate: float = 0.001,
                 num_workers: int = 4,
                 train_ratio: float = 0.8):
        """
        Specialized trainer for image super resolution tasks
        
        Args:
            model_name (str): Name of the model to initialize
            dataset_paths (List[str]): Paths to datasets
            batch_size (int): Batch size for training
            learning_rate (float): Learning rate for optimizer
            num_workers (int): Number of workers for data loading
            train_ratio (float): Ratio of data to use for training (rest goes to validation)
        """
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.dataset_paths = dataset_paths
        self.model_name = model_name
        self.learning_rate = learning_rate
        self.train_ratio = train_ratio
        
        # Initialize datasets and loaders
        self._initialize_datasets()
        
        # Initialize model and training parameters
        self._initialize_model()
    
    def _initialize_datasets(self):
        """
        Helper method to initialize datasets
        """
        # Read training data based on input format
        if isinstance(self.dataset_paths, list):
            df_train = pd.concat([pd.read_parquet(path) for path in self.dataset_paths], ignore_index=True)
        else:
            raise ValueError("Invalid dataset_paths format. Must be a list or dictionary.")
        
        df_train, df_val = train_test_split(
            df_train,
            test_size=1 - self.train_ratio,
            random_state=42
        )
        
        # Define preprocessing transforms
        self.transform = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])
        ])
        
        # Create datasets and dataloaders
        self.train_dataset = ImageDataset(df_train, transform=self.transform)
        self.val_dataset = ImageDataset(df_val, transform=self.transform)
        
        self.train_loader = DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers
        )
        
        self.val_loader = DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers
        ) if df_val is not None else None
    
    def _initialize_model(self):
        """
        Helper method to initialize model architecture and training parameters
        """
        # Load base model
        self.model = AIIABase.load(self.model_name)
        
        # Freeze CNN layers
        for param in self.model.cnn.parameters():
            param.requires_grad = False
            
        # Add upscaling layer
        hidden_size = self.model.config.hidden_size
        kernel_size = self.model.config.kernel_size
        self.model.upsample = nn.Sequential(
            nn.Upsample(scale_factor=2, mode='bilinear', align_corners=False),
            nn.Conv2d(hidden_size, 3, kernel_size=kernel_size, padding=1)
        )
        
        # Initialize optimizer and loss function
        self.criterion = nn.MSELoss()
        self.optimizer = torch.optim.Adam(
            [param for param in self.model.parameters() if 'upsample' in str(param)],
            lr=self.learning_rate
        )
        
        self.best_val_loss = float('inf')
    
    def train(self, num_epochs: int = 10):
        """
        Train the model for specified number of epochs
        """
        self.model.to(self.device)
        
        for epoch in range(num_epochs):
            print(f"Epoch {epoch+1}/{num_epochs}")
            
            # Train phase
            self._train_epoch()
            
            # Validation phase
            if self.val_loader is not None:
                self._validate_epoch()
                
            # Save best model based on validation loss
            if self.val_loader is not None and self.current_val_loss < self.best_val_loss:
                self.model.save("aiuNN-base")
    
    def _train_epoch(self):
        """
        Train model for one epoch
        """
        self.model.train()
        running_loss = 0.0
        
        for batch in self.train_loader:
            low_res = batch['low_ress'].to(self.device)
            high_res = batch['high_ress'].to(self.device)

            # Forward pass
            features = self.model.cnn(low_res)
            outputs = self.model.upsample(features)

            loss = self.criterion(outputs, high_res)

            # Backward pass and optimize
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

            running_loss += loss.item()
        
        epoch_loss = running_loss / len(self.train_loader)
        print(f"Train Loss: {epoch_loss:.4f}")
    
    def _validate_epoch(self):
        """
        Validate model performance
        """
        self.model.eval()
        val_oss = 0.0

        with torch.no_grad():
            for batch in self.val_loader:
                low_res = batch['low_ress'].to(self.device)
                high_res = batch['high_ress'].to(self.device)

                features = self.model.cnn(low_res)
                outputs = self.model.upsample(features)

                loss = self.criterion(outputs, high_res)
                val_loss += loss.item()

        avg_val_loss = val_loss / len(self.val_loader) if self.val_loader else 0
        print(f"Validation Loss: {avg_val_loss:.4f}")
        
        # Update best model
        if avg_val_loss < self.best_val_loss:
            self.best_val_loss = avg_val_loss
    
    def __repr__(self):
        return f"Model ({self.model_name}, batch_size={self.batch_size})"
    
if __name__ == "__main__":
    trainer = ModelTrainer(
        model_name="/root/vision/AIIA/AIIA-base-512/",
        dataset_paths=[
            "/root/training_data/vision-dataset/image_upscaler.parquet",
            "/root/training_data/vision-dataset/image_vec_upscaler.parquet"
        ],
        batch_size=2,
        learning_rate=0.001
    )
    
    trainer.train(num__epochs=3)